; curl.egg
; Chicken Scheme bindings for libcurl.
;
; Copyright (c) 2007, Tony Sidaway
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are
; permitted provided that the following conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of
;   conditions and the following disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list
;   of conditions and the following disclaimer in the documentation and/or other materials
;   provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse
;   or promote products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
; MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
; THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
; OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
; HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
; TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
; EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to: 
;
; tonysidaway@gmail.com
;
; Alternatives:
; * Report bugs and suggestions on Chicken trac:
; ** http://trac.callcc.org/
; * Subscribe to the mailing list chicken-users@nongnu.org to discuss this egg.
; ** http://lists.nongnu.org/mailman/listinfo/chicken-users
;
(use tinyclos) ; (Gregor Kiczales, adapted for Chicken by Felix Winkelmann)
(use lolevel)  ; For object-evict.
(use srfi-1)   ; List Library (Olin Shivers)
(use srfi-4)   ; Homogeneous numeric vector datatypes (Marc Feeley)
(use srfi-69)  ; Basic hash tables (Panu Kalliokoski) 
#>
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>
/* C_char *new_c_string (C_word s)
 * Copy a Scheme string object s to a newly allocated C string.
 * If given a non-string object, or if allocation fails, this
 * function returns a NULL pointer.
 */
static C_char *new_c_string (C_word scheme_string) {
C_char *string_field, *c_string;
C_word string_length;
    /* Return NULL if we're given a non-string object */
    if ((scheme_string & C_IMMEDIATE_MARK_BITS) ||
	(C_header_bits(scheme_string) != C_STRING_TYPE)) {
        return NULL;
    }
    string_length = C_header_size(scheme_string);
    c_string = (C_char *)C_malloc(string_length+1);
    /* Don't try to populate if the allocation fails */
    if (NULL != c_string) {
	string_field = C_c_string(scheme_string);
        C_memcpy (c_string, string_field, string_length);
	c_string[string_length] = '\0';
    }
    return c_string;
}
/*
 * Wrapper for C pointers of various kinds.
 */
typedef struct {
  void *ptr;
} c_object;
<#
(define new-c-string (foreign-lambda c-pointer "new_c_string" scheme-object))
;
(define-foreign-record (c-object c_object)
  (constructor: make-c-object)
  (destructor: delete-c-object)
  (c-pointer ptr))
;
; Error codes:
;
(define curlcodes '(
  0 ok                                              ;0
  unsupported-protocol failed-init  url-malformat  url-malformat-user  couldnt-resolve-proxy
  couldnt-resolve-host  couldnt-connect  ftp-weird-server-reply  ftp-access-denied
  ftp-user-password-incorrect                     ;10
  ftp-weird-pass-reply  ftp-weird-user-reply  ftp-weird-pasv-reply  ftp-weird-227-format
  ftp-cant-get-host  ftp-cant-reconnect  ftp-couldnt-set-binary  partial-file
  ftp-couldnt-retr-file  ftp-write-error          ;20
  ftp-quote-error  http-returned-error  write-error  malformat-user  ftp-couldnt-stor-file
  read-error  out-of-memory  operation-timeouted  ftp-couldnt-set-ascii
  ftp-port-failed                                 ;30
  ftp-couldnt-use-rest  ftp-couldnt-get-size  http-range-error  http-post-error
  ssl-connect-error  bad-download-resume  file-couldnt-read-file  ldap-cannot-bind
  ldap-search-failed  library-not-found           ;40
  function-not-found  aborted-by-callback  bad-function-argument  bad-calling-order
  interface-failed  bad-password-entered  too-many-redirects  unknown-telnet-option
  telnet-option-syntax  obsolete                  ;50
  ssl-peer-certificate  got-nothing  ssl-engine-notfound  ssl-engine-setfailed
  send-error  recv-error  share-in-use  ssl-certproblem  ssl-cipher
  ssl-cacert                                      ;60
  bad-content-encoding  ldap-invalid-url  filesize-exceeded  ftp-ssl-failed
  send-fail-rewind  ssl-engine-initfailed  login-denied  tftp-notfound
  tftp-perm  tftp-diskfull                        ;70
  tftp-illegal  tftp-unknownid  tftp-exists  tftp-nosuchuser))
;
(define curlmcodes '(
  1 call-multi-perform ok bad-handle bad-easy-handle out-of-memory internal-error))
;
(define curlshcodes '(
  0 ok  bad-option in-use invalid nomem nomem))
;
; (curl-error-string errorcode)
; Translates a numeric or symbolic error code into a string.
;
(define (curl-error-string error-code error-list error-function)
  (let* ((error-number (cond
     ((symbol? error-code) (- (length error-list) (length (or (memq error-code (cdr error-list)) '())) (car error-list) 1))
     ((number? error-code) error-code)
     (else -1000)))
     (error-c-string (error-function error-number))
     (len ((foreign-lambda size-t "strlen" c-string) error-c-string))
     (error-string (make-string len)))
    (move-memory! error-c-string error-string len)
    error-string))
;
; (filter-index pred lst)
; Filter a list according to a predicate on the list index
;
(define (filter-index pred lst)
  (let ((ilst (iota (length lst))))
    (map (lambda (i) (list-ref lst i)) (filter pred ilst))))
;
(define port-hash #f) ;; Hash table for storage of ports used in callbacks.
;
; Some handy foreign types
;
(define-foreign-type size-t unsigned-long)
(define-foreign-type curl-pointer c-pointer)
(define-foreign-type curl-code int)
(define-foreign-type file-pointer c-pointer)
(define-foreign-type curl-option int)
;
; This is the minimum number of characters that an errorbuffer must hold.
;
(define curl-error-size (foreign-value "CURL_ERROR_SIZE" int))
;
; (curl_writefunc_port ptr size nmemb port-number)
; Callback function that sends nmemb items of size bytes to the Scheme
; port designated by port-number
;
(define-external
  (curl_writefunc_port ; c function name
   ; arguments:
   (c-pointer ptr) (size-t size)(size-t nmemb) (unsigned-long port-number))
  size-t ; return type
    (let* ((nbytes (* size nmemb))
	 (s (make-string nbytes))
	 (port-thunk (hash-table-ref port-hash port-number)))
      (move-memory! ptr s nbytes)
      (display s (port-thunk))
      nbytes))
;
; Some foreign declarations of C functions from libcurl which are used by
; this Scheme extension.
;
(define curl-easy-init
  (foreign-safe-lambda curl-pointer "curl_easy_init"))
(define curl-easy-perform
  (foreign-safe-lambda curl-code "curl_easy_perform" curl-pointer))
(define curl-easy-set-option-long!
  (foreign-safe-lambda curl-code "curl_easy_setopt" curl-pointer curl-option long))
(define curl-easy-set-option-c-string!
  (foreign-safe-lambda curl-code "curl_easy_setopt" curl-pointer curl-option c-string))
(define curl-easy-set-option-function!
  (foreign-safe-lambda curl-code "curl_easy_setopt" curl-pointer curl-option c-pointer))
(define curl-easy-set-option-c-pointer!
  (foreign-safe-lambda curl-code "curl_easy_setopt" curl-pointer curl-option c-pointer))
(define curl-easy-set-option-file!
  (foreign-safe-lambda curl-code "curl_easy_setopt" curl-pointer curl-option file-pointer))
(define curl-easy-cleanup!
  (foreign-safe-lambda void "curl_easy_cleanup" curl-pointer))
(define curl-easy-reset!
  (foreign-safe-lambda void "curl_easy_reset" curl-pointer))
(define curl-slist-append
  (foreign-safe-lambda c-pointer "curl_slist_append" c-pointer c-string))
;
(define curl-share-init
  (foreign-safe-lambda curl-pointer "curl_share_init"))
(define curl-share-set-option-long!
  (foreign-safe-lambda int "curl_share_setopt" curl-pointer curl-option long))
(define curl-share-set-option-c-string!
  (foreign-safe-lambda int "curl_share_setopt" curl-pointer curl-option c-string))
(define curl-share-set-option-function!
  (foreign-safe-lambda int "curl_share_setopt" curl-pointer curl-option c-pointer))
(define curl-share-set-option-c-pointer!
  (foreign-safe-lambda int "curl_share_setopt" curl-pointer curl-option c-pointer))
(define curl-share-set-option-file!
  (foreign-safe-lambda int "curl_share_setopt" curl-pointer curl-option file-pointer))
(define curl-share-cleanup!
  (foreign-safe-lambda void "curl_share_cleanup" curl-pointer))

(define curl-multi-init
  (foreign-safe-lambda curl-pointer "curl_multi_init"))
;;(define curl-multi-set-option-long!
;;  (foreign-safe-lambda curl-code "curl_multi_setopt" curl-pointer curl-option long))
;;(define curl-multi-set-option-c-pointer!
;;  (foreign-safe-lambda curl-code "curl_multi_setopt" curl-pointer curl-option c-pointer))
;
;========================================================================
;
; (curl-set-scheme-port c port)
; Store a Scheme port in the port-hash table (SRFI-69) and return the key, a port
; number.
(define curl-set-scheme-port!
  (let ((port-number -1))
    (lambda (c port)
      (let ((real-curl (or (slot-ref c '%real-curl) (curl-easy-init))))
	(slot-set! c '%real-curl real-curl)
	(unless port-hash (set! port-hash (make-hash-table =)))
	(set! port-number (+ port-number 1))
	; There has to be a limit, though it doesn't need to be a low one.
	(and (> port-number 32767) (set! port-number 0))
	(hash-table-set! port-hash port-number port)
	port-number))))
;
;========================================================================
;
; <curl>: Base type for all curl types.
;
(define-class <curl>
  (<object>)
  (%type %c-objects))
;
; (curl-finalize! c)
; Dispose of unmanaged resources that pertain to a <curl>
; instance.
;
(define-method (curl-finalize! (c <curl>))
  (map free (slot-ref c '%c-objects))
  (slot-set! c '%c-objects '()))
;
;========================================================================
;
; <curl-easy>
;
;
; Options for <curl-easy>
; Option names preceded by "_" are for the most part Schemified versions of the
; "real" options.  The option name denotes a scheme expression which is transformed
; in someway to produce the "_" version, which is suitable for handling by the
; C code.
; See curl-easy-special-options.
;
(define curl-easy-options
    `(
    ; (name type curl-option-value)
    (_write long ,(foreign-value "CURLOPT_FILE" int))
    (_read long ,(foreign-value "CURLOPT_INFILE" int))
    (_header long ,(foreign-value "CURLOPT_HEADERDATA" int))
    (_stderr file ,(foreign-value "CURLOPT_STDERR" int))
    (_sslversion long ,(foreign-value "CURLOPT_SSLVERSION" int)) 
    (_http_version long ,(foreign-value "CURLOPT_HTTP_VERSION" int))
    (_httpauth long ,(foreign-value "CURLOPT_HTTPAUTH" int))
    (_proxyauth long ,(foreign-value "CURLOPT_PROXYAUTH" int))
    (_ipresolve long ,(foreign-value "CURLOPT_IPRESOLVE" int)) 
    (_ftp_ssl long ,(foreign-value "CURLOPT_FTP_SSL" int)) 
    (_ftpsslauth long ,(foreign-value "CURLOPT_FTPSSLAUTH" int)) 
    (url c-string ,(foreign-value "CURLOPT_URL" int))
    (port long ,(foreign-value "CURLOPT_PORT" int))
    (proxy  c-string ,(foreign-value "CURLOPT_PROXY" int))
    (userpwd c-string ,(foreign-value "CURLOPT_USERPWD" int))
    (proxyuserpwd string ,(foreign-value "CURLOPT_PROXYUSERPWD" int)) 
    (range c-string ,(foreign-value "CURLOPT_RANGE" int)) 
    (writefunction function ,(foreign-value "CURLOPT_WRITEFUNCTION" int)) 
    (readfunction function ,(foreign-value "CURLOPT_READFUNCTION" int)) 
    (timeout long ,(foreign-value "CURLOPT_TIMEOUT" int))
    (infilesize long ,(foreign-value "CURLOPT_INFILESIZE" int)) 
    (postfields c-string ,(foreign-value "CURLOPT_POSTFIELDS" int)) 
    (referer c-string ,(foreign-value "CURLOPT_REFERER" int))
    (ftpport c-string ,(foreign-value "CURLOPT_FTPPORT" int)) 
    (useragent c-string ,(foreign-value "CURLOPT_USERAGENT" int)) 
    (low_speed_limit long ,(foreign-value "CURLOPT_LOW_SPEED_LIMIT" int))
    (low_speed_time long ,(foreign-value "CURLOPT_LOW_SPEED_TIME" int)) 
    (resume_from long ,(foreign-value "CURLOPT_RESUME_FROM" int))
    (cookie c-string ,(foreign-value "CURLOPT_COOKIE" int))
    (httpheader s-list ,(foreign-value "CURLOPT_HTTPHEADER" int)) 
    (httppost s-list ,(foreign-value "CURLOPT_HTTPPOST" int)) 
    (sslcert c-string ,(foreign-value "CURLOPT_SSLCERT" int))
    (sslcertpasswd c-string ,(foreign-value "CURLOPT_SSLCERTPASSWD" int)) 
    (sslkeypasswd c-string ,(foreign-value "CURLOPT_SSLKEYPASSWD" int)) 
    (crlf long ,(foreign-value "CURLOPT_CRLF" int))
    (quote s-list ,(foreign-value "CURLOPT_QUOTE" int)) 
    (cookiefile c-string ,(foreign-value "CURLOPT_COOKIEFILE" int))
    (timecondition long ,(foreign-value "CURLOPT_TIMECONDITION" int)) 
    (timevalue long ,(foreign-value "CURLOPT_TIMEVALUE" int))
    (customrequest c-string ,(foreign-value "CURLOPT_CUSTOMREQUEST" int)) 
    (postquote s-list ,(foreign-value "CURLOPT_POSTQUOTE" int))
    (writeinfo c-string ,(foreign-value "CURLOPT_WRITEINFO" int)) 
    (verbose long ,(foreign-value "CURLOPT_VERBOSE" int)) 
    (header long ,(foreign-value "CURLOPT_HEADER" int)) 
    (noprogress long ,(foreign-value "CURLOPT_NOPROGRESS" int))
    (nobody long ,(foreign-value "CURLOPT_NOBODY" int))
    (failonerror long ,(foreign-value "CURLOPT_FAILONERROR" int)) 
    (upload long ,(foreign-value "CURLOPT_UPLOAD" int)) 
    (post long ,(foreign-value "CURLOPT_POST" int))
    (ftplistonly long ,(foreign-value "CURLOPT_FTPLISTONLY" int))
    (ftpappend long ,(foreign-value "CURLOPT_FTPAPPEND" int)) 
    (netrc long ,(foreign-value "CURLOPT_NETRC" int))
    (followlocation long ,(foreign-value "CURLOPT_FOLLOWLOCATION" int)) 
    (transfertext long ,(foreign-value "CURLOPT_TRANSFERTEXT" int)) 
    (put long ,(foreign-value "CURLOPT_PUT" int))
    (progressfunction function ,(foreign-value "CURLOPT_PROGRESSFUNCTION" int)) 
    (progressdata c-pointer ,(foreign-value "CURLOPT_PROGRESSDATA" int))
    (autoreferer long ,(foreign-value "CURLOPT_AUTOREFERER" int)) 
    (proxyport long ,(foreign-value "CURLOPT_PROXYPORT" int))
    (postfieldsize long ,(foreign-value "CURLOPT_POSTFIELDSIZE" int))
    (httpproxytunnel long ,(foreign-value "CURLOPT_HTTPPROXYTUNNEL" int)) 
    (interface c-string ,(foreign-value "CURLOPT_INTERFACE" int)) 
    (krb4level c-string ,(foreign-value "CURLOPT_KRB4LEVEL" int))
    (ssl_verifypeer long ,(foreign-value "CURLOPT_SSL_VERIFYPEER" int)) 
    (cainfo c-string ,(foreign-value "CURLOPT_CAINFO" int)) 
    (maxredirs long ,(foreign-value "CURLOPT_MAXREDIRS" int))
    (filetime long ,(foreign-value "CURLOPT_FILETIME" int)) 
    (telnetoptions s-list ,(foreign-value "CURLOPT_TELNETOPTIONS" int)) 
    (maxconnects long ,(foreign-value "CURLOPT_MAXCONNECTS" int))
    (fresh_connect long ,(foreign-value "CURLOPT_FRESH_CONNECT" int))
    (forbid_reuse long ,(foreign-value "CURLOPT_FORBID_REUSE" int)) 
    (random_file c-string ,(foreign-value "CURLOPT_RANDOM_FILE" int))
    (egdsocket c-string ,(foreign-value "CURLOPT_EGDSOCKET" int)) 
    (connecttimeout long ,(foreign-value "CURLOPT_CONNECTTIMEOUT" int)) 
    (headerfunction function ,(foreign-value "CURLOPT_HEADERFUNCTION" int))
    (httpget long ,(foreign-value "CURLOPT_HTTPGET" int)) 
    (ssl_verifyhost long ,(foreign-value "CURLOPT_SSL_VERIFYHOST" int)) 
    (cookiejar c-string ,(foreign-value "CURLOPT_COOKIEJAR" int))
    (ssl_cipher_list c-string ,(foreign-value "CURLOPT_SSL_CIPHER_LIST" int)) 
    (ftp_use_epsv long ,(foreign-value "CURLOPT_FTP_USE_EPSV" int))
    (sslcerttype c-string ,(foreign-value "CURLOPT_SSLCERTTYPE" int)) 
    (sslkey c-string ,(foreign-value "CURLOPT_SSLKEY" int)) 
    (sslkeytype c-string ,(foreign-value "CURLOPT_SSLKEYTYPE" int))
    (sslengine c-string ,(foreign-value "CURLOPT_SSLENGINE" int)) 
    (sslengine_default c-string ,(foreign-value "CURLOPT_SSLENGINE_DEFAULT" int))
    (dns_use_global_cache long ,(foreign-value "CURLOPT_DNS_USE_GLOBAL_CACHE" int)) 
    (dns_cache_timeout long ,(foreign-value "CURLOPT_DNS_CACHE_TIMEOUT" int)) 
    (prequote s-list ,(foreign-value "CURLOPT_PREQUOTE" int))
    (debugfunction function ,(foreign-value "CURLOPT_DEBUGFUNCTION" int)) 
    (debugdata c-pointer ,(foreign-value "CURLOPT_DEBUGDATA" int))
    (cookiesession long ,(foreign-value "CURLOPT_COOKIESESSION" int)) 
    (capath c-string ,(foreign-value "CURLOPT_CAPATH" int)) 
    (buffersize long ,(foreign-value "CURLOPT_BUFFERSIZE" int))
    (nosignal long ,(foreign-value "CURLOPT_NOSIGNAL" int)) 
    (share curl-share ,(foreign-value "CURLOPT_SHARE" int)) 
    (proxytype long ,(foreign-value "CURLOPT_PROXYTYPE" int)) 
    (encoding c-string ,(foreign-value "CURLOPT_ENCODING" int))
    (private c-string ,(foreign-value "CURLOPT_PRIVATE" int)) 
    (http200aliases s-list ,(foreign-value "CURLOPT_HTTP200ALIASES" int))
    (unrestricted_auth long ,(foreign-value "CURLOPT_UNRESTRICTED_AUTH" int))
    (ftp_use_eprt long ,(foreign-value "CURLOPT_FTP_USE_EPRT" int)) 
    (ssl_ctx_function function ,(foreign-value "CURLOPT_SSL_CTX_FUNCTION" int)) 
    (ssl_ctx_data c-pointer ,(foreign-value "CURLOPT_SSL_CTX_DATA" int))
    (ftp_create_missing_dirs long ,(foreign-value "CURLOPT_FTP_CREATE_MISSING_DIRS" int)) 
    (ftp_response_timeout long ,(foreign-value "CURLOPT_FTP_RESPONSE_TIMEOUT" int)) 
    (maxfilesize long ,(foreign-value "CURLOPT_MAXFILESIZE" int))
    (infilesize_large curl-offset ,(foreign-value "CURLOPT_INFILESIZE" int)) 
    (resume_from_large curl-offset ,(foreign-value "CURLOPT_RESUME_FROM_LARGE" int))
    (maxfilesize_large curl-offset ,(foreign-value "CURLOPT_MAXFILESIZE_LARGE" int)) 
    (netrc_file c-string ,(foreign-value "CURLOPT_NETRC_FILE" int))
    (postfieldsize_large curl-offset ,(foreign-value "CURLOPT_POSTFIELDSIZE_LARGE" int))
    (tcp_nodelay long ,(foreign-value "CURLOPT_TCP_NODELAY" int)) 
    (source_userpwd c-string ,(foreign-value "CURLOPT_USERPWD" int))
    (source_prequote s-list ,(foreign-value "CURLOPT_PREQUOTE" int)) 
    (source_postquote s-list ,(foreign-value "CURLOPT_POSTQUOTE" int))
    (ioctlfunction function ,(foreign-value "CURLOPT_IOCTLFUNCTION" int)) 
    (ioctldata c-pointer ,(foreign-value "CURLOPT_IOCTLDATA" int))
    (ftp_account c-string ,(foreign-value "CURLOPT_FTP_ACCOUNT" int))
    (cookielist c-string ,(foreign-value "CURLOPT_COOKIELIST" int)) 
    (ignore_content_length long ,(foreign-value "CURLOPT_IGNORE_CONTENT_LENGTH" int))
    (ftp_skip_pasv_ip long ,(foreign-value "CURLOPT_FTP_SKIP_PASV_IP" int))
    (ftp_filemethod long ,(foreign-value "CURLOPT_FTP_FILEMETHOD" int))))
;
; List of curl options in the easy interface that have to be handled
; specially.
; _write, _read, and _header are versions of the curl write/read options
; that are able to perform Scheme input-output.
; _stderr is an "odd man out" because output to stderr is handled directly by
; libcurl (there is no option to change the error output function).
; So we only handle it specially in order to give an error message if
; the programmer tries to redirect curl stderr to a scheme port.
; ftp_ssl, ftl-sslauth, http_version, httpauth, proxyauth, ipresolve,
; and sslversion have Schemified input forms that have to be handled
; in a special way.
; errorbuffer is handled in a unique way.  The actual value given for this
; option is ignored except to see if it evaluates to #f.  If not, subsequent
; invocations of (curl-get-option c 'errorbuffer) return the last error
; message (or #f if there has been none so far).  If it does evaluate to #f,
; subsequent invocations of (curl-get-option c 'errorbuffer) always return #f.
;
(define curl-easy-special-options
    '((file  _write #t) (writedata  _write #t) (infile _read #t)
      (readdata _read #t) (writeheader  _header #t) (headerdata _header #t)
      (stderr  _stderr #t) (ftp_ssl _ftp_ssl #f) (ftpsslauth _ftpsslauth #f)
      (http_version _http_version #f) (httpauth _httpauth #f) (proxyauth _proxyauth #f)
      (ipresolve _ipresolve #f) (sslversion _sslversion #f) (errorbuffer errorbuffer #f)))
;
; (curl-easy-set-option! c rc name value)
; Translate a <curl-easy> option into something C will understand, and
; set it.  Returns value argument it was given.
;
(define (curl-easy-set-option! c rc name value)
  (let ((option-record (assq name curl-easy-options)))
    ; Test option lookup for validity.
    (when (not (and option-record
	 (pair? option-record) (not (null? option-record))
	 (pair? (cdr option-record)) (not (null? (cdr option-record)))))
	 (error (format "Bad option name \"~S\" for <curl-easy>" name)))
    (let* ((type (cadr option-record))
	  (key (caddr option-record))
	  (return-code
	   (case type
	     ((file) (curl-easy-set-option-file! rc key value))
	     ((long) (curl-easy-set-option-long! rc key (if (number? value) value (if value 1 0))))
	     ((c-string) (curl-easy-set-option-c-pointer! rc key
							  (curl-make-c-string c name)))
	     ((function) (curl-easy-set-option-c-pointer! rc key (c-object-ptr value)))
	     ((curl-share) (curl-easy-set-option-c-pointer! rc key (slot-ref value '%real-curl)))
	     ((s-list)
	      (or (pair? value) (error "Non-list option when s-list is required."))
	      (let ((s-list (make-slist c)))
		(c-object-ptr-set! s-list (null-pointer))
		(let loop ((value value))
		  (if (null? value)
		      (curl-easy-set-option-c-pointer! rc key (c-object-ptr s-list))
		      (begin
			(or (string? (car value)) (error "Non-string in list parameter"))
			(c-object-ptr-set! s-list (curl-slist-append (c-object-ptr s-list) (car value)))
			(loop (cdr value)))))))
	     (else (error (format
			   "unsupported <curl-easy> option type \"~S\"" type))))))
      (if (>= return-code (length curlcodes))
	  'unknown-error
	  (list-ref curlcodes return-code)))))

;
; <curl-easy>: class to handle libcurl's easy interface
;
(define <curl-easy>
  (make-class
    (list <curl>)
    (append '(%real-curl %errorbuffer errorbuffer ftp_ssl ftp_sslauth http_version httpauth proxyauth ipresolve sslversion)(map car curl-easy-options))))


;
(slot-set! <curl-easy> 'name 'curl-easy)
;
; (curl-set-option! (c <curl-easy>) name value)
; Basic setter for <curl-easy>
; Returns the value it was given.
;
(define-method (make-slist  (c <curl-easy>))
  (let ((s (make-c-object)))
    (slot-set! c '%c-objects (cons s (slot-ref c '%c-objects)))
    s))
;
; (curl-set-option! (c <curl-easy>) name value)
; Basic setter for <curl-easy>
; Returns the value it was given.
;
(define-method (curl-set-option!  (c <curl-easy>) name value)
  (let ((special (assq name curl-easy-special-options)))
    (if special
	(case (cadr special)
	  ((errorbuffer)
	   (slot-set! c 'errorbuffer value))
	  ((_stderr)
	   (and (or (port? value) (procedure? value))
		(error
		 "Must specify a C file pointer for <curl-easy> stderr, not a port thunk or port")
	   (slot-set! c (cadr special) value)))
	  ((_http_version)
	   (or (not value) (symbol? value) (equal? value 1.0) (equal? value 1.1)
	       (error "http_version must be a symbol, or 1.0 or 1.1"))
	   (case value
	     ((#f) (slot-set! c (cadr special) #f))
	     ((none)
	      (slot-set! c (cadr special) (foreign-value "CURL_HTTP_VERSION_NONE" long)))
	     ((1.0 1_0)
	      (slot-set! c (cadr special) (foreign-value "CURL_HTTP_VERSION_1_0" long)))
	     ((1.1 1_1)
	      (slot-set! c (cadr special) (foreign-value "CURL_HTTP_VERSION_1_1" long)))
	     (else
	      (error (format "Bad <curl-easy> http_version value ~S" value))))
	   (slot-set! c 'http_version value)) ; for display
	  ((_ftp_ssl)
	   (or (not value) (symbol? value) (error "ftp_ssl must be a symbol"))
	   (case value
	     ((#f) (slot-set! c (cadr special) #f))
	     ((none)
	      (slot-set! c (cadr special) (foreign-value "CURLFTPSSL_NONE" long)))
	     ((try)
	      (slot-set! c (cadr special) (foreign-value "CURLFTPSSL_TRY" long)))
	     ((control)
	      (slot-set! c (cadr special) (foreign-value "CURLFTPSSL_CONTROL" long)))
	     ((all)
	      (slot-set! c (cadr special) (foreign-value "CURLFTPSSL_ALL" long)))
	     (else
	      (error (format "Bad <curl-easy> ftp_ssl value ~S" value))))
	   (slot-set! c 'ftp_ssl value)) ; for display
	  ((_ftp_sslauth)
	   (or (not value) (symbol? value) (error "ftp_sslauth must be a symbol"))
	   (case value
	     ((#f) (slot-set! c (cadr special) #f))
	     ((default)
	      (slot-set! c (cadr special) (foreign-value "CURLFTPAUTH_DEFAULT" long)))
	     ((ssl)
	      (slot-set! c (cadr special) (foreign-value "CURLFTPAUTH_SSL" long)))
	     ((tls)
	      (slot-set! c (cadr special) (foreign-value "CURLFTPAUTH_TLS" long)))
	     (else
	      (error (format "Bad <curl-easy> ftp_sslauth value ~S" value))))
	   (slot-set! c 'ftp_sslauth value)) ; for display
	  ((_timecondition)
	   (or (not value) (symbol? value) (error "timecondition must be a symbol"))
	   (case value
	     ((#f) (slot-set! c (cadr special) #f))
	     ((ifmodsince)
	      (slot-set! c (cadr special) (foreign-value "CURL_TIMECOND_IFMODSINCE" long)))
	     ((ifunmodsince)
	      (slot-set! c (cadr special) (foreign-value "CURL_TIMECOND_IFUNMODSINCE" long)))
	     (else
	      (error (format "Bad <curl-easy> timecondition value ~S" value))))
	   (slot-set! c 'timecondition value)) ; for display
	  ((_ipresolve)
	   (or (not value) (symbol? value) (error "ipresolve must be a symbol"))
	   (case value
	     ((#f) (slot-set! c (cadr special) #f))
	     ((whatever)
	      (slot-set! c (cadr special) (foreign-value "CURL_IPRESOLVE_WHATEVER" long)))
	     ((v4 ipv4)
	      (slot-set! c (cadr special) (foreign-value "CURL_IPRESOLVE_V4" long)))
	     ((v6 ipv6)
	      (slot-set! c (cadr special) (foreign-value "CURL_IPRESOLVE_V6" long)))
	     (else
	      (error (format "Bad <curl-easy> ipresolve value ~S" value))))
	   (slot-set! c 'ipresolve value)) ; for display
	  ((_sslversion)
	   (or (not value) (symbol? value) (error "sslversion must be a symbol"))
	   (case value
	     ((#f) (slot-set! c (cadr special) #f))
	     ((default)
	      (slot-set! c (cadr special) (foreign-value "CURL_SSLVERSION_DEFAULT" long)))
	     ((tlsv1)
	      (slot-set! c (cadr special) (foreign-value "CURL_SSLVERSION_TLSv1" long)))
	     ((sslv2)
	      (slot-set! c (cadr special) (foreign-value "CURL_SSLVERSION_SSLv2" long)))
	     ((sslv3)
	      (slot-set! c (cadr special) (foreign-value "CURL_SSLVERSION_SSLv3" long)))
	     (else
	      (error (format "Bad <curl-easy> sslversion value ~S" value))))
	   (slot-set! c 'sslversion value)) ; for display
	  ((_httpauth _proxyauth)
	   (if (symbol? value) (set! value (list value))) ;; NB: MUTATED VALUE!
	   (or (not value) (list? value) (error (format "<curl-easy> ~S option must be a symbol or list of symbols" name)))
	   (slot-set! c (cadr special)
	     (if (not value)
	       #f
	       (let loop ((option 0)(lst value))
		 (if (null? lst)
		     option
		     (case (car lst)
		       ((basic) (loop (bitwise-ior option (foreign-value "CURLAUTH_BASIC" long)) (cdr lst)))
		       ((digest) (loop (bitwise-ior option (foreign-value "CURLAUTH_DIGEST" long)) (cdr lst)))
		       ((gssnegotiate) (loop (bitwise-ior option (foreign-value "CURLAUTH_GSSNEGOTIATE" long)) (cdr lst)))
		       ((ntlm) (loop (bitwise-ior option (foreign-value "CURLAUTH_NTLM" long)) (cdr lst)))
		       ((any) (loop (bitwise-ior option (foreign-value "CURLAUTH_ANY" long)) (cdr lst)))
		       (else (error (format "Bad <curl-easy> ~S option ~S" name (car lst)))))))))
	   (slot-set! c (car special) value)) ; for display
	  ((_write _read _header)
	   (or (procedure? value)
	       (error (format
		       "Must specify a thunk evaluating to a port for <curl-easy> ~S option" name)))
	   (slot-set! c (cadr special)
		      (and value (curl-set-scheme-port! c value)))))
	(begin
	  (slot-set! c name value))))
  value)
;
; (curl-set-options! (c <curl-easy>) . rest)
; Compound setter for <curl-easy>: handles many sets in one go.
; Returns the first value it was given.
;
(define-method (curl-set-options!  (c <curl-easy>) . rest)
  (and (not (null? rest))
       (car (map (lambda (x y)(curl-set-option! c x y))
		 (filter-index even? rest)
		 (filter-index odd? rest)))))
;
; (send (c <curl-easy>) . rest)
; Shorthand function for various operations.
;
(define-method (send  (c <curl-easy>) . rest)
  (cond
   ((null? rest) (map (lambda (x)(cons x (slot-ref c x))) (map car (class-slots <curl-easy>))))
   ((equal? '(perform) rest)(curl-perform c))
   ((= 1 (length rest))(curl-get-option c (car rest)))
   (else (apply curl-set-options! c rest))))
;
; (curl-get-option (c <curl-easy>) name)
; Basic getter for <curl-easy>
;
(define-method (curl-get-option  (c <curl-easy>) name)
  (if (eq? name 'errorbuffer)
    (let ((errormessage (slot-ref c '%errorbuffer)))
      (if (and (slot-ref c 'errorbuffer) errormessage)
	(let* ((length ((foreign-lambda size-t "strlen" c-pointer) errormessage))
	       (errorbuffer (make-string length)))
	  (move-memory! errormessage errorbuffer length)
	  errorbuffer)
	#f))
    (let ((special (assq name curl-easy-special-options)))
      (if (and special (caddr special))
	  (slot-ref c (cadr special))
	  (slot-ref c name)))))
;
; (curl-reset! (c <curl-easy>))
; Reset a <curl-easy> object to the initial state
;
(define-method (curl-reset!  (c <curl-easy>) . release)
  (let ((real-curl (or (curl-get-option c '%real-curl)
		       (curl-set-option! c '%real-curl (curl-easy-init))))
	(opt-list (filter
		   (lambda (x)
		     (and
		      (not (char=? (string-ref (symbol->string x) 0) #\%))
		      (curl-get-option c x)))
		   (map car (class-slots <curl-easy>)))))
    ; Remove stale hash table entries
    (curl-easy-remove-port-hash-entries! c)
    ; Remove error buffer if it exists.
    (let ((errorbuffer (slot-ref c '%errorbuffer)))
      (if errorbuffer
	  (begin
	    (free errorbuffer)
	    (slot-set! c '%errorbuffer #f))))
    (slot-set! c 'errorbuffer #f)
    (map (lambda (name) (slot-set! c name #f)) opt-list)
    (curl-set-options! c
		       'file current-output-port
		       'writefunction curl-writefunc-port
		       'stderr #f
		       'writeheader #f
		       'infile #f)
    (if (and release (pair? release) (car release))
	(curl-release! c)
	(curl-easy-reset! real-curl))
    (map free (slot-ref c '%c-objects))
    (slot-set! c '%c-objects '())))
;
; (curl-release! (c <curl-easy>))
; Release the real curl, if there is one.
;
(define-method (curl-release! (c <curl-easy>))
  (let ((real-curl (slot-ref c '%real-curl)))
    (if real-curl
	(begin
	  (curl-easy-cleanup! real-curl)
	  (slot-set! c '%real-curl #f)))))

;
; (curl-set-error-buffer! (c <curl-easy>) rc value)
; if value is #f, clears any error buffer allocated, otherwise
; makes sure an error buffer is allocated.
;
(define-method (curl-set-error-buffer! (c <curl-easy>) rc value)
  ; Ignore any actual error buffer string argument and create an internal
  ; error buffer which will be used instead.
  (let ((real-error-buffer (slot-ref c '%errorbuffer))) ; NB: THIS VALUE MUTATES
    (if (not value) ; Make sure that any allocated buffer is freed.
      (if real-error-buffer
	  (begin (free real-error-buffer)
		 (set! real-error-buffer (null-pointer)) ; NB: MUTATED VALUE
		 (slot-set! c '%errorbuffer #f)))
      (if (not real-error-buffer)
	  (begin
	    (set! real-error-buffer
		  (new-c-string (make-string (+ 1 curl-error-size))))
	    (slot-set! c '%errorbuffer real-error-buffer)))) ; NB: MUTATED VALUE
    (curl-easy-set-option-c-pointer! rc (foreign-value "CURLOPT_ERRORBUFFER" int)
				     real-error-buffer)))
;
; (curl-perform (c <curl-easy>))
; Translates all <curl-easy> option values that have been set by the program
; and applies it to a curl, then calls curl_easy_perform().
;
(define-method (curl-perform (c <curl-easy>))
  (let ((real-curl (or (curl-get-option c '%real-curl)
		       (curl-set-option! c '%real-curl (curl-easy-init))))
	(opt-list (lset-difference eq?
		    (filter
		     (lambda (x)
		       (and
			(not (char=? (string-ref (symbol->string x) 0) #\%))
			(curl-get-option c x)))
		     (map car (class-slots <curl-easy>)))
		    (map car curl-easy-special-options))))
    (map (lambda (x)
	   (curl-easy-set-option! c real-curl x (curl-get-option c x)))
	 opt-list)
    (curl-set-error-buffer! c real-curl (slot-ref c 'errorbuffer))
    (let ((return-code (curl-easy-perform real-curl)))
      (if (>= return-code (+ (length curlcodes) (car curlcodes)))
	  'unknown-error
	  (list-ref curlcodes (+ return-code 1 (car curlcodes)))))))
;
; (curl-make-c-string (c <curl-easy>) name)
; Translates the string value in a named <curl-easy> slot into
; a c string, and returns the c string.
;
(define-method (curl-make-c-string (c <curl>) name)
  (let* ((obj (slot-ref c name))
	(c-string (new-c-string obj)))
    (slot-set! c '%c-objects (cons c-string (slot-ref c '%c-objects)))
    c-string))
;
; (initialize (c <curl>))
; Initializer for <curl> objects.
;
(add-method initialize
  (make-method (list <curl>)
    (lambda (cnm obj initargs)
      (slot-set! obj '%c-objects '()))))
;
; (initialize (c <curl-easy>))
; Initializer for <curl-easy> objects.
; Handles a bunch of initializer pairs for <curl-easy>
;
(add-method initialize
  (make-method (list <curl-easy>)
    (lambda (cnm obj initargs)
      (map (lambda (x) (curl-set-option! obj (car x) #f))
	   (class-slots <curl-easy>))
      (curl-set-option! obj '%type 'easy)
      (slot-set! obj '%c-objects '())
      (apply curl-set-options! obj initargs))))
;
; (curl-easy-remove-port-hash-entries! c)
; Remove port-hash entries that pertain to a curl-easy instance.
;
(define-method (curl-easy-remove-port-hash-entries! (c <curl>))
  (let ((write  (slot-ref c '_write))
	(header (slot-ref c '_header))
	(read (slot-ref c '_read)))
    (when write (hash-table-delete! port-hash write))
    (when header (hash-table-delete! port-hash header))
    (when read (hash-table-delete! port-hash read))))
;
; (curl-easy-finalize! c)
; Dispose of unmanaged resources that pertain to a <curl-easy>
; instance.

(define-method (curl-easy-finalize! (c <curl-easy>))
  (curl-easy-remove-port-hash-entries! c)
  (let ((errorbuffer (slot-ref c '%errorbuffer)))
    (if errorbuffer
	(begin
	  (free errorbuffer)
	  (slot-set! c '%errorbuffer #f))))
  (let ((real-curl (slot-ref c '%real-curl)))
    (when real-curl (curl-easy-cleanup! real-curl)))
  (curl-finalize! c))
;
; (curl-easy-error-string errorcode)
; Translates a numeric or symbolic error code into a string.
;
(define (curl-easy-error-string error-code)
  (curl-error-string error-code curlcodes (foreign-lambda c-string "curl_easy_strerror" int)));
;
;
;========================================================================
;
; <curl-multi>
; TODO: implement this class.
;
;
; List of options for <curl-multi>
;
(define curl-multi-options
    '((socketfunction . callback) (socketdata . pointer) (pipelining . int)
      (timefunction . callback) (timerdata . pointer)))
;
; <curl-multi>: class to handle libcurl's multi interface
;
(define <curl-multi>
  (make-class
   (list <curl>)
   (list
    '%real-curl 'socketfunction 'socketdata 'pipelining 'timefunction 'timerdata)))
;
(slot-set! <curl-multi> 'name 'curl-multi)
;
; (curl-set-option! (c <curl-multi>) name value)
; Basic setter for <curl-multi>
; Returns the value it was given.
;
(define-method (curl-set-option!  (c <curl-multi>) name value)
  (or (symbol? name) (error "curl-set-option (<curl-multi>): ~S is not a symbol"))
  (let ((real-curl (or (slot-ref c '%real-curl) (curl-init-multi))))
    (slot-set c '%real-curl real-curl)
    (case name
      ((socketfunction timefunction socketdata timerdata)
       (curl-multi-set-option-c-pointer! rc name value))
      ((pipelining) (slot-set! c name value)(curl-multi-set-option-long! rc name value))
      (else (error (format "curl-set-option! (<curl-multi>): option ~S not recognised" name))))))
;
; (curl-set-options! (c <curl-multi>) . rest)
; Compound setter for <curl-multi>: handles many sets in one go.
; Returns the first value it was given.
;
(define-method (curl-set-options!  (c <curl-multi>) . rest)
  (display "(curl-set-options! <curl-multi>) is not yet implemented.")(newline))
;
; (curl-perform (c <curl-multi>))
; Translates all <curl-multi> option values that have been set by the program
; and applies it to a curl, then calls curl_multi_perform().
;
(define-method (curl-perform (c <curl-multi>))
  (let ((real-curl (or (curl-get-option c '%real-curl)
		       (curl-set-option! c '%real-curl (curl-multi-init))))
	(opt-list  (filter
		     (lambda (x)
		       (and
			(not (char=? (string-ref (symbol->string x) 0) #\%))
			(curl-get-option c x)))
		     (map car (class-slots <curl-multi>)))))
;;    (map (lambda (x)
;;	   (curl-multi-set-option! c real-curl x (curl-get-option c x)))
;;	 opt-list)
    (let ((return-code (curl-multi-perform real-curl)))
      (if (>= return-code (+ (length curlmcodes) (car curlmcodes)))
	  'unknown-error
	  (list-ref curlmcodes (+ return-code 1 (car curlmcodes)))))))
;
; (initialize (c <curl-multi>))
; This handles a bunch of initializers for <curl-multi>
;
(add-method initialize
  (make-method (list <curl-multi>)
    (lambda (cnm obj initargs)
      (slot-set! obj '%type 'multi)
      (map (lambda (x) (curl-set-option! obj (car x) #f))
	   (class-slots <curl-multi>))
      (map (lambda (x y)(curl-set-option! obj x y))
	   (filter-index even? initargs)
	   (filter-index odd? initargs)))))
;
; (curl-multi-error-string errorcode)
; Translates a numeric or symbolic error code into a string.
;
(define (curl-multi-error-string error-code)
  (curl-error-string error-code curlmcodes (foreign-lambda c-string "curl_multi_strerror" int)));
;
;
;========================================================================
;
; <curl-share>
;
; <curl-share>: class to handle libcurl's share interface
;
(define <curl-share>
  (make-class
   (list <curl>)
   (list
    '%real-curl 'lockfunc 'unlockfunc 'userdata 'share-cookies 'share-dns)))
;
(slot-set! <curl-share> 'name 'curl-share)
;
; (curl-share-set-option! c rc name value)
; Translate a <curl-share> option into something C will understand, and
; set it.  Returns value argument it was given.
;
(define (curl-share-set-option! c rc key type value)
  (case type
    ((long) (curl-share-set-option-long! rc key (if (number? value) value (if value 1 0))) value)
    ((function c-pointer) (curl-share-set-option-c-pointer! rc key (c-object-ptr value))
     value)
    (else (error (format
		  "unsupported <curl-share> option type \"~S\"" type)))))
;
; (curl-set-option! (c <curl-share>) name value)
; Basic setter for <curl-share>
; Returns the value it was given.
;
(define-method (curl-set-option!  (c <curl-share>) name value)
  (let ((real-curl (or (curl-get-option c '%real-curl)
		       (let ((rc (curl-share-init))) (slot-set! c '%real-curl rc) rc)))
	(lookup (assq name `(
	  (lockfunc ,(foreign-value "CURLSHOPT_LOCKFUNC" int))
	  (unlockfunc ,(foreign-value "CURLSHOPT_UNLOCKFUNC" int))
	  (share ,(foreign-value "CURLSHOPT_SHARE" int))
	  (unshare ,(foreign-value "CURLSHOPT_UNSHARE" int))
	  (userdata ,(foreign-value "CURLSHOPT_USERDATA" int))))))
    (case name
      ((lockfunc unlockfunc)
       (slot-set! c name value)
       (curl-share-set-option! c real-curl (cadr lookup) 'function value)
       value)
      ((userdata)
       (slot-set! c name value)
       (curl-share-set-option-c-pointer! c real-curl (cadr lookup) 'c-pointer value)
       value)
      ((share unshare)
       (let ((lst (if (pair? value) value (list value)))
	     (currentop (cadr lookup))
	     (otherop (if (eq? name 'share) (foreign-value "CURLSHOPT_UNSHARE" int)
		      (foreign-value "CURLSHOPT_SHARE" int))))
	 (let loop ((lst lst))
	   (if (null? lst)
	       #f
	       (begin
		 (case (car lst)
		   ((dns) (let ((share-dns (slot-ref c 'share-dns)))
			    (if (eqv? (eq? name 'unshare) share-dns) ; Is a change required?
			      (begin
				(slot-set! c 'share-dns (not share-dns))
				(curl-share-set-option! c real-curl currentop 'long
				  (foreign-value "CURL_LOCK_DATA_DNS" long))))))
		   ((cookies) (let ((share-cookies (slot-ref c 'share-cookies)))
				(if (eqv? (eq? name 'unshare) share-cookies) ; Is a change required?
				  (begin
				    (slot-set! c 'share-cookies (not share-cookies))
				    (curl-share-set-option! c real-curl currentop 'long
				(foreign-value "CURL_LOCK_DATA_COOKIE" long))))))
		   (else (error (format "Bad <curl-share> share value ~S" (car lst)))))
		 (loop (cdr lst))))))
       value)
      (else (error (format "Bad <curl-share> option ~S" name))))))
;
; (curl-set-options! (c <curl-share>) . rest)
; Compound setter for <curl-share>: handles many sets in one go.
; Returns the first value it was given.
;
(define-method (curl-set-options!  (c <curl-share>) . rest)
  (and (not (null? rest))
       (car (map (lambda (x y)(curl-set-option! c x y))
		 (filter-index even? rest)
		 (filter-index odd? rest)))))
;
; (curl-get-option (c <curl-easy>) name)
; Basic getter for <curl-easy>
;
(define-method (curl-get-option  (c <curl-share>) name)
  (slot-ref c name))
;
; (initialize (c <curl-share>))
; This initializes <curl-share> to a null state.
;
(add-method initialize
  (make-method (list <curl-share>)
    (lambda (cnm obj initargs)
      (slot-set! obj '%type 'share)
      (slot-set! obj '%real-curl #f)
      (slot-set! obj 'lockfunc #f)
      (slot-set! obj 'unlockfunc #f)
      (slot-set! obj 'userdata #f)
      (slot-set! obj 'share-dns #f)
      (slot-set! obj 'share-cookies #f))))
;
; (curl-perform (c <curl-share>))
; curl-perform on a <curl-share> is an error.
;
(define-method (curl-perform (c <curl-share>))
  (error "Cannot curl-perform a <curl-share>!"))
;
; (curl-share-error-string errorcode)
; Translates a numeric or symbolic error code into a string.
;
(define (curl-share-error-string error-code)
  (curl-error-string error-code curlshcodes (foreign-lambda c-string "curl_share_strerror" int)));
;
;========================================================================
;
; Generic creator for <curl-easy>, <curl-share> and <curl-multi>
;
;
; (curl type . rest)
; A curl creator.  Creates a curl of the required type, and
; applies initial options to it.
(define (curl type . rest)
  (case type
    ((easy)
     (let ((c (make <curl-easy>)))
       (curl-set-options! c
			  'file current-output-port
			  'writefunction curl-writefunc-port)
       (apply curl-set-options! c rest)
       (set-finalizer! c curl-easy-finalize!)
       c))
    ((multi)
     (apply make <curl-multi> rest))
    ((share)
     (apply make <curl-share> rest))
    (else
     (error "curl type must be 'easy, 'share or 'multi"))))
;
;========================================================================
;
; Module initialization code
;
; The C function curl_global_init must be called before libcurl can be
; used.
((foreign-safe-lambda void "curl_global_init" long) (foreign-value "CURL_GLOBAL_ALL" long))
;
;
; We don't need to use curl_writefunc_port from Scheme, but we do need to pass
; its address in <curl-easy> options, and so give it a scheme-like name.
;
(define curl-writefunc-port (let ((f (make-c-object)))(c-object-ptr-set! f (foreign-value "curl_writefunc_port" c-pointer)) f))
;
; END